;;; $Id: natScheme2.scm,v 1.6 1996/09/25 10:49:30 queinnec Exp $

;;; Une variante qui cree les variables inexistantes au niveau global.

(define (update!! id env value)
  (if (pair? env)
      (if (eq? (caar env) id)
          (begin (set-cdr! (car env) value)
                 value )
          (update!! id (cdr env) value) )
      (begin (set-cdr! (last-pair env.global)
                       (list (cons id value)) )
             value ) ) )

(set! update! update!!)

(define (scheme2)
  (display "Bienvenue !")
  (newline)
  (let toplevel ()
    (display "?? ")
    (let ((e (expand-program (read))))
      (if (eof-object? e)
          (begin (display ";;; end")
                 (newline) )
          (let ((r (evaluate e env.global)))
            (display "== ")
            (display r)
            (newline)
            (toplevel) ) ) ) ) )

(define (expand-program e)
  (evaluate `(expand ',e) env.predefined) )

(define (expand-expression e)
  (if (pair? e)
      (case (car e)
        ((quote) e)
        ((lambda) `(lambda ,(cadr e) . ,(expand-expressions (cddr e))))
        (else 
         (let ((expander (assoc (car e) macro-env)))
           (if (pair? expander)
               ;; Utiliser le protocole d'appel du Scheme interpr\'et\'e.
               (let ((ee (invoke (cdr expander) (list e))))
                 (expand ee) )
               (expand-expressions e) ) ) ) )
      e ) )

(define (expand-expressions e*)
  (if (pair? e*)
      (cons (expand-expression (car e*))
            (expand-expressions (cdr e*)) )
      e* ) )

(define (macro-definer e)
  (let* ((call (cadr e))
         (body (cddr e))
         (name (car call))
         (vars (cdr call)) )
    ;; {\tt macro-env} est partag\'e avec {\tt expand-expression}.
    (set! macro-env
          ;; Il faut \'evaluer l'expanseur pour qu'il puisse \^etre invoqu\'e
          (cons (cons name (evaluate `(lambda (e) 
                                        (apply (lambda ,vars . ,body) 
                                               (cdr e) ) )
                                     env.predefined ) )
                macro-env ) )
    `(quote ,name) ) )

(define macro-env 
  (list (cons 'define-macro macro-definer)) )

;;; representation differente des fonctions.

(define (make-function variables body env)
  (lambda values
    (eprogn body (extend env variables values)) ) )

(define (invoke fn args)
  (if (procedure? fn)
      (apply fn args)
      (wrong "Not a function" fn) ) )

;;; Reset the environment.

(set! env.init '())

;;; Three macros to define the initial global environment

(set! env.global env.init)

;;; Ici la verification d'arite est laissee a l'implantation.

(define-macro (defprimitive name value arity)
  `(definitial ,name ,value) )

(define-macro (defnaryprimitive name value arity)
  `(definitial ,name ,value) )

(define-macro (defpredicate name value arity)
  `(definitial ,name ,value) )

(defprimitive cons cons 2)
(defprimitive car car 1)
(defprimitive cdr cdr 1)
(defpredicate pair? pair? 1)
(defpredicate symbol? symbol? 1)
(defpredicate eq? eq? 2)
(defprimitive set-car! set-car! 2)
(defprimitive set-cdr! set-cdr! 2)
(defnaryprimitive + + 2)
(defprimitive - - 2 )
(defprimitive = = 2)
(defpredicate < < 2)
(defpredicate > > 2)
(defnaryprimitive * * 2)
(defpredicate <= <= 2)
(defpredicate >= >= 2)
(defprimitive remainder remainder 2)

(defprimitive call/cc 
   (lambda (f) 
     (call/cc (lambda (g) 
                (invoke
                 f (list (lambda values
                           (if (= (length values) 1)
                               (g (car values))
                               (wrong "Incorrect arity" g) ) ))) )) )
   1 )

(defnaryprimitive apply apply 2)

(defnaryprimitive list list 0)

(defprimitive display display 1)
(defprimitive newline newline 0)

;;; Some fundamental constants:
(definitial t #t)
(definitial f #f)
(definitial nil '())

;;; Some free global locations:
(definitial x 'void)
(definitial y 'void)
(definitial z 'void)
(definitial a 'void)
(definitial b 'void)
(definitial c 'void)
(definitial foo 'void)
(definitial bar 'void)
(definitial hux 'void)
(definitial fib 'void)
(definitial fact 'void)
(definitial visit 'void)
(definitial length 'void)
(definitial primes 'void)

(set! env.predefined
      (cons (cons 'expand expand-expression)
            ;; copie l'environnement global afin d'\^etre ind\'ependant.
            (append env.global '()) ) )
                  
;;; Since now global variables are automatically created skip that tests:

(set! tests-to-skip 
      '(
        (set! xyzzy 3)
        ) )

;;; Redefine that function in Bigloo to get underlying exceptions.

(define (test-natscheme file)
  (suite-test file "NatScheme2? " "NatScheme2= " 
     #t
     (lambda (read check error)
       (set! wrong error)
       (lambda ()   
         (define (skip-read)
           (let ((e (read)))
             (if (member e tests-to-skip)
                 (begin (read)              ; skip the associated result
                        (skip-read) )
                 e ) ) )
         (check (try (evaluate (expand-program (skip-read)) env.global)
                     (lambda (k a b c)
                       (wrong `(error ,a ,b ,c)) ) )) ) )
     equal? ) )
 
;;; end of natScheme2.scm
